#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <linux/types.h>
#include <linux/spi/spidev.h>

#define LEN 20
#define SPEED_HZ 500000 //5000 kHz (75 max)


/* Fonction générique de transmission SPI */

int transSPI(int fd, void * tx, void * rx,int len)
{
	int status;
	struct spi_ioc_transfer xfer={
		.tx_buf = (unsigned long) tx,
		.rx_buf = (unsigned long) rx,
		.len = len,
		.speed_hz=SPEED_HZ,
		.delay_usecs=0,
		.bits_per_word=8,
		.cs_change=1
	};

	status = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (status < 0) {
		perror("Erreur\nSPI_IOC_MESSAGE");
		return 0;
	}

	return 0;
}

/*  Initialisation de spidev */
int initSPI()
{
	int fd; 

	fd = open("/dev/spidev1.0", O_RDWR);
	if (fd < 0) {
		perror("Erreur\nOpen spidevice ");
		return 1;
	}
	return fd;
}

/* Lecture de la description de l'EEPROM 
 * Doit être toujours 20h 20h 14h */
int lectureDesc(int fd)
{
	int i;
	 /* Ajout de la commande de lecture */
	unsigned char command=0x9F,buffer[4];

	transSPI(fd,(void *)&command,(void *)buffer,4);

	printf("response ");
	for (i=0; i<5;i++)
		printf(" %02x", buffer[i]);
	printf("\n");

	return 0;
}

/* Lecture de l'EEPROM, stockage du résultat dans un buffer */
int lecture(int fd,unsigned long add,unsigned char * buffer,int len){

	 /* Ajout de la commande de lecture */
 
	unsigned char txbuff[4];
	txbuff[0]=0x03;
	txbuff[1] = (add & 0x00FF0000)>> 16;
	txbuff[2] = (add & 0x0000FF00)>> 8;
	txbuff[3] = (add & 0x000000FF);
	
    transSPI(fd,(void *)txbuffer,(void *)buffer,len);

	return 0;
}

/* Activation du bit WEL (Write enable latch) */
int writeEnable(int fd)
{
	unsigned char we=0x06;
	transSPI(fd,(void *)&we,NULL,1);
	return 0;
}

/* Lecture du registre d'état */
unsigned char readReg(int fd)
{
	unsigned char command=0x05,reg[2];
	transSPI(fd,(void *)&command,(void *)reg,2);
	return reg[1];
}

/* Fonction d'écriture d'un fichier dans l'EEPROM */
int ecritureBitstream(int fd,FILE * bitstream)
{
	int status,nb=0,nbtot=0;
	struct spi_ioc_transfer xfer;
	/* Buffer de transmission : 256 octets pour les donn?es,
	 * 3 pour l'adresse et 1 pour la commande d'ecriture */
	unsigned char buffer[256+4];
	buffer[0]=0x02;
	buffer[1]=0x00;
	buffer[2]=0x00;
	buffer[3]=0x00;
	xfer.rx_buf = (unsigned long) NULL;
	xfer.speed_hz=SPEED_HZ;
	xfer.delay_usecs=0;
	xfer.bits_per_word=8;
	xfer.cs_change=1;
	xfer.tx_buf = (unsigned long) buffer;
	xfer.len=260;

	/* Transmission des données */
	while((nb=fread(&(buffer[4]),1,256,bitstream))>0 && buffer[1]<0x0F )
	{
		nbtot+=nb;
		writeEnable(fd);
		status = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
		if (status < 0) {
			perror("Erreur\nSPI_IOC_MESSAGE");
			return 0;
		}
#ifdef DEBUG
        /* Si le bit de WIP n'est pas à 1 */
		if(readReg(fd)!=3)
			printf("Erreur\n Erreur d'écriture a l'adresse 0x%02x%02x%02x\n",buffer[1],buffer[2],buffer[3]);
#endif
        /* Attente de la fin de l'écriture */
		while((readReg(fd)&0x01)==1){}

        /* Incrémentation de l'adresse */
		if(buffer[2]==0xFF)
		{
			buffer[1]++;
			buffer[2]=0;
		}
		if(buffer[2]!=0)
			buffer[2]++;
		else
			buffer[2]=0x01;

	}
	return nbtot;
}

/* Écriture du registre de status  */
int ecrireReg(int fd,unsigned char val)
{
	unsigned char command[2]={0x01,val};
	transSPI(fd,(void *)command,(void *)NULL,2);
	return 0;
}

/* Écriture d'un fichier à partir de la mémoire */
int ecritureFichier(int fd,int fichier,int size)
{
	unsigned long adresse=0x000000;
	unsigned char buffer[4096];
while(size>4092)
{
	lecture(fd,adresse,buffer,4096);
	size-=4092;
	adresse+=0x001000;
	write(fichier,buffer+4,4092);
}
	lecture(fd,adresse,buffer,size+4);
	write(fichier,buffer+4,size);
	return 0;
}

int main(int argc, char ** argv)
{
	// Récupération du nom de fichier passé en argument 

	static struct option long_option =  {"upload",1,NULL,'u'};
	int c=getopt_long(argc,argv,"u:",&long_option,NULL);
	const char * file=NULL;
	if(c=='u'){
		file=optarg;
	}
	else
	{
			printf("usage : ./lecture --upload FICHIER ou -u FICHIER\n");
			exit(-1);
	}

	// Initialisation de Spidev 
	int fd=initSPI();

	// Bulk erase 
	unsigned char be=0xC7;
	writeEnable(fd);
	transSPI(fd,(void *)&be,(void *)NULL,1);

#ifdef DEBUG
	printf("reg : %d\n",readReg(fd));
	printf("reg : %d\n",readReg(fd));
	lectureDesc(fd);
#endif
	// Attente que la mémoire soit prète 
	while((readReg(fd)&0x01)==1)
	{
#ifdef DEBUG
		printf("effacement en cours\n");
		sleep(1);
#endif
	}

	int nb;
	FILE * fichier=fopen(file,"r");
	if(fichier!=NULL)
		nb=ecritureBitstream(fd,fichier);
	else
		printf("Erreur\nEchec de l'ouverture du fichier\n");

#ifdef DEBUG
	printf("nb = %d\n",nb); 
#endif

	// Effacement de l'ancien bitstream.test
	remove("/www/cgi-bin/bitstream.test");

	//Ecriture du fichier contenant le contenu de la mémoire
	int fdtest=open("/www/cgi-bin/bitstream.test",O_RDWR | O_CREAT, S_IRUSR | S_IRGRP | S_IROTH);
	ecritureFichier(fd,fdtest,nb);
	return 0;
}

